@@ -123,13 +123,20 @@ getExpression.CallrFuture <- function(future, mc.cores = 1L, ...) {
 #' @importFrom future resolved
 #' @keywords internal
 #' @export
-resolved.CallrFuture <- function(x, ...) {
+resolved.CallrFuture <- function(x, .signalEarly = TRUE, ...) {
   resolved <- NextMethod()
   if (resolved) return(TRUE)
   
   process <- x$process
   if (!inherits(process, "r_process")) return(FALSE)
-  !process$is_alive()
+  resolved <- !process$is_alive()
+
+  ## Signal errors early?
+  if (.signalEarly && resolved) {
+    ## Trigger a FutureError already here, if exit code != 0
+    if (process$get_exit_status() != 0L) result(x)
+  }
+  resolved
 }
 
 #' @importFrom future result UnexpectedFutureResultError
@@ -167,9 +174,6 @@ result.CallrFuture <- function(future, ...) {
 #' @S3method run CallrFuture
 #' @export
 run.CallrFuture <- local({
-  FutureRegistry <- import_future("FutureRegistry")
-  assertOwner <- import_future("assertOwner")
-
   ## MEMOIZATION
   cmdargs <- NULL
 
@@ -260,8 +264,6 @@ run.CallrFuture <- local({
 #' @importFrom utils tail
 #' @importFrom future FutureError FutureWarning
 await <- local({
-  FutureRegistry <- import_future("FutureRegistry")
-
   function(future, timeout = getOption("future.wait.timeout", 30*24*60*60),
                    delta = getOption("future.wait.interval", 1.0),
                    alpha = getOption("future.wait.alpha", 1.01),
@@ -331,7 +333,8 @@ await <- local({
     
     ## Failed?
     if (inherits(result, "error")) {
-      stop(CallrFutureError(result, future = future))
+      msg <- post_mortem_failure(result, future = future)
+      stop(CallrFutureError(msg, future = future))
     }
     
     if (debug) mdebugf("- callr:::get_result() ... done (after %d attempts)", ii)
@@ -374,3 +377,55 @@ await <- local({
     result
   } # await()
 })
+
+
+
+post_mortem_failure <- function(reason, future) {
+  assert_no_references <- import_future("assert_no_references")
+  summarize_size_of_globals <- import_future("summarize_size_of_globals")
+
+  stop_if_not(inherits(future, "Future"))
+  
+  ## (1) Trimmed error message
+  if (inherits(reason, "error")) reason <- conditionMessage(reason)
+
+  ## (2) Information on the future
+  label <- future$label
+  if (is.null(label)) label <- "<none>"
+  stop_if_not(length(label) == 1L)
+
+  ## (3) POST-MORTEM ANALYSIS:
+  postmortem <- list()
+                 
+  process <- future$process
+  pid <- tryCatch(process$get_pid(), error = function(e) NA_integer_)
+  start_time <- tryCatch(format(process$get_start_time(), format = "%Y-%m-%dT%H:%M:%S%z"), error = function(e) NA_character_)
+  msg2 <- sprintf("The parallel worker (PID %.0f) started at %s", pid, start_time)
+  if (process$is_alive()) {
+    msg2 <- sprintf("%s is still running", msg2)
+  } else {
+    exit_code <- tryCatch(process$get_exit_status(), error = function(e) NA_integer_)
+    msg2 <- sprintf("%s finished with exit code %.0f", msg2, exit_code)
+  }
+  postmortem$alive <- msg2
+
+  ## (c) Any non-exportable globals?
+  globals <- future[["globals"]]
+  postmortem$non_exportable <- assert_no_references(globals, action = "string")
+
+  ## (d) Size of globals
+  postmortem$global_sizes <- summarize_size_of_globals(globals)
+
+  ## (4) The final error message
+  msg <- sprintf("%s (%s) failed. The reason reported was %s",
+                 class(future)[1], label, sQuote(reason))
+  stop_if_not(length(msg) == 1L)
+  if (length(postmortem) > 0) {
+    postmortem <- unlist(postmortem, use.names = FALSE)
+    msg <- sprintf("%s. Post-mortem diagnostic: %s",
+                   msg, paste(postmortem, collapse = ". "))
+    stop_if_not(length(msg) == 1L)
+  }
+
+  msg
+} # post_mortem_failure()
